/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2011                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#ifndef MPILINKTEST_OUTPUT_SION_PAR_H
#define MPILINKTEST_OUTPUT_SION_PAR_H
int linktest_output_sion_parallel_mpi ( _linktest_spec *linktest_spec, _linktest_stat *linktest_stat );
int linktest_output_sion_2stage_mpi ( _linktest_spec *linktest_spec, _linktest_stat *linktest_stat );
#endif
